from .base import BaseProtocol, BaseHoneypot

class Socks5Protocol(BaseProtocol):
    '''based on RFC 1928 (https://datatracker.ietf.org/doc/html/rfc1928) '''

    SOCKS_VERSION = b'\x05'
    METHOD = b'\x02' # USERNAME/PASSWORD
       
    def __init__(self, config=None):
        if config:
            self.config = config
        self.protocol_name = "socks5"

#    def connection_made(self, transport):
#        self.transport = transport
#        self.log_data("connection_made", self.transport.get_extra_info('peername'), time.time())
#        
#    def data_received(self, data):    
#        self.log_data("data_received", self.transport.get_extra_info('peername'), time.time(), {'data':data})
#
#        if data[:1] == b'\x05':
#            self.initial_handshake(data)
#        elif data[:1] == b'\x01':
#            self.authenticate(data)
#        else:
#            self.socks_failed()
#
#    def initial_handshake(self, payload):
#        try:
#            #+----+----------+----------+
#            #|VER | NMETHODS | METHODS  |
#            #+----+----------+----------+
#            #| 1  |    1     | 1 to 255 |
#            #+----+----------+----------+
#            ver, nmethods, methods  = struct.unpack("!BB%ds" % (len(payload)-2), payload)
#        except struct.error:
#            # not a socks handshake packet
#            self.socks_failed()
#
#        if ver != 0x05:
#            # version not correct
#            self.socks_failed()
#
#        if self.METHOD not in methods:
#            # client doesn't support password
#            self.socks_failed()
#
#        self.transport.write(self.SOCKS_VERSION + self.METHOD)
#
#    def authenticate(self, payload):
#        username_length = payload[1]
#        username = payload[2:2 + username_length]
#        password_length = int.from_bytes( payload[2 + username_length:2 + username_length + 1], "little")
#        
#        if password_length != 0:
#            password = payload[-password_length:]
#        else:
#            password = b""
#        
#        self.log_data("authenticate", self.transport.get_extra_info('peername'), time.time(), {'username': username,'password':password})
#        self.socks_failed()
#        return username, password
#
#    def socks_failed(self):
#        self.transport.write(self.SOCKS_VERSION + b'\xFF') # NO ACCEPTABLE METHODS
#        self.connection_lost(None)

class Socks5Honeypot(BaseHoneypot):

   def __init__(self, config, logger, bindaddr="0.0.0.0"):
        super().__init__(config, logger, bindaddr)
        self.handler = lambda: Socks5Protocol(config=config)
        self.handler.logger = logger
        self.handler.config = config


from .base import BaseProtocol, BaseHoneypot

# Telnet command definitions
IAC  = b'\xff'  # Interpret as Command
WILL = b'\xfb'
WONT = b'\xfc'
DO   = b'\xfd'
DONT = b'\xfe'
SE   = b'\xf0'  # End of subnegotiation parameters
NOP  = b'\xf1'  # No operation
DM   = b'\xf2'  # Data mark
BRK  = b'\xf3'  # Break
IP   = b'\xf4'  # Interrupt process
AO   = b'\xf5'  # Abort output
AYT  = b'\xf6'  # Are you there
EC   = b'\xf7'  # Erase character
EL   = b'\xf8'  # Erase line
GA   = b'\xf9'  # Go ahead
SB   = b'\xfa'  # Subnegotiation of the indicated option follows

class TelnetProtocol(BaseProtocol):
    """
    Telnet Honeypot server based on https://www.rfc-editor.org/rfc/rfc854
    """
    
    versions = {
        "D-Link DSL router": b"\xff\xfd\x01\xff\xfd!\xff\xfb\x01\xff\xfb\x03\r\nlogin: ",
    }

    def __init__(self, config=None):
        if config:
            self.config = config
        self.protocol_name = "telnet"
        self.username = b''
        self.password = b''
        self.state = 'USERNAME'

    def connection_made(self, transport):
        self.transport = transport
        self.logger.log(self.protocol_name + "." + self.logger.CONNECTION, self.transport)

        self.transport.write(self.versions[self.config['version']])

    def data_received(self, data):

        if IAC in data:
            self.logger.log(self.protocol_name + "." + self.logger.DATA, self.transport, data=data)
            self.handle_telnet_command(data)
        else:
            self.handle_user_input(data)

    def handle_telnet_command(self, data):
        # Handle incoming Telnet commands here
        i = 0
        while i < len(data):
            if data[i] == IAC[0]:
                i += 1
                if i < len(data) and data[i] in [WILL[0], WONT[0], DO[0], DONT[0]]:
                    option = data[i]
                    i += 1
                    if i < len(data):
                        option_code = data[i]
                        # For simplicity, we'll just respond with DONT for DO and WONT for WILL
                        if option == DO[0]:
                            self.transport.write(IAC + DONT + bytes([option_code]))
                        elif option == WILL[0]:
                            self.transport.write(IAC + WONT + bytes([option_code]))
                        i += 1
                else:
                    # Handle other Telnet commands like SE, NOP, etc.
                    i += 1
            else:
                i += 1

    def handle_user_input(self, data):
        if data == b"\x7f":  # DEL character
            self.erase_character()
        
        if self.state == 'USERNAME':
            if b'\n' in data or b'\r' in data:
                self.transport.write(b"\r\nPassword: ")
                self.state = 'PASSWORD'
            else:
                self.username += data
                self.transport.write(data)
        elif self.state == 'PASSWORD':
            if b'\n' in data or b'\r' in data:
                self.authenticate()
            else:
                self.password += data

    def erase_character(self):
        # Intentionally not removing data from self.username or self.password, to keep informations
        # Send backspace to client
        self.transport.write(b'\b \b')

    def authenticate(self):
        username = self.username.decode('utf-8', errors='replace')
        password = self.password.decode('utf-8', errors='replace')
        self.logger.log(self.protocol_name + "." + self.logger.LOGIN, self.transport, extra={"username": username, "password": password})
        self.transport.write(b"\r\nLogin incorrect.\r\n")
        self.transport.close()

class TelnetHoneypot(BaseHoneypot):

    def __init__(self, config, logger, bindaddr="0.0.0.0"):
        super().__init__(config, logger, bindaddr)
        self.handler = lambda: TelnetProtocol(config=config)
        self.handler.logger = logger
        self.handler.config = config



import asyncio
import psutil
import argparse, json, socket, os

from . import __version__
from .modules import *
from .logger import *

class TrapsterManager:
    def __init__(self, config):
        self.logger = None
        self.config = config

    def get_ip(self, config_interface):
        for interface, addrs in psutil.net_if_addrs().items():
            if interface == config_interface:
                for addr in addrs:
                    if addr.family == socket.AF_INET:
                        return addr.address
        
        print(f"Interface {config_interface} does not exist, using 0.0.0.0")
        return

    async def start(self):
        ip = self.get_ip(self.config['interface'])

        for service_type in self.config['services']:
            for service_config in self.config['services'][service_type]:

                if service_type == 'ftp':
                    server = FtpHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'http':
                    server = HttpHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'https':
                    server = HttpsHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'ssh':
                    server = SshHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'dns':            
                    server = DnsHoneypot(service_config, self.logger, bindaddr=ip, proxy_dns_ip="127.0.0.1")
                elif service_type == 'vnc':
                    server = VncHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'mysql':
                    server = MysqlHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'postgres':
                    server = PostgresHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'ldap':
                    server = LdapHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'rdp':
                    server = RdpHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'telnet':
                    server = TelnetHoneypot(service_config, self.logger, bindaddr=ip)
                elif service_type == 'snmp':
                    server = SnmpHoneypot(service_config, self.logger, bindaddr=ip)
                else:
                    print(f"[-] Unreconized service {service_type}")
                    break
                                        
                try:
                    print(f"[+] Starting service {service_type} on port {service_config['port']}")
                    await server.start()
                except Exception as e:
                    print(e)
        
        while True:
            await asyncio.sleep(10)

def list_interfaces():
    interfaces = psutil.net_if_addrs()
    for interface, addrs in interfaces.items():
        for addr in addrs:
            if addr.family == socket.AF_INET:
                print(f"{addr.address}\t({interface})")

def main():
    parser = argparse.ArgumentParser(description="Trapster Community honeypot.")
    
    parser.add_argument('-i', '--interfaces', action='store_true', help='Show list of interfaces and their corresponding IPs.')
    parser.add_argument('-c', '--config', type=str, help='Specify the config file to use.')
    parser.add_argument('-s', '--show-config', action='store_true', help='Show the config file currently in use.')
    parser.add_argument('-v', '--version', action='store_true', help='Print version')
    args = parser.parse_args()
    
    if args.version:
        print(__version__)
        return

    if args.interfaces:
        list_interfaces()
        return
    
    if args.config:
        config_file = args.config
        print(f"[+] using config file at : {config_file}")
    else:
        config_file = os.path.dirname(__file__)+"/data/trapster.conf"
        print(f"[+] using default config file at: {config_file}")
    
    if os.path.exists(config_file):
        with open(config_file, 'r') as f:
            config = json.load(f)
    else:
        print(f'[-] config file {config_file} not found')
        return

    if args.show_config:
        print(config)
        return

    print('=== Starting Trapster Community ===')
    manager = TrapsterManager(config)

    logger = JsonLogger(config['id'])
    logger.whitelist_ips = []

    manager.logger = logger

    try:
        asyncio.run(manager.start())
    except KeyboardInterrupt:
        print('Finishing')



from .base import BaseProtocol, BaseHoneypot

class FtpProtocol(BaseProtocol):

    config = {
        'banner': "Microsoft FTP Service"
       # 'passwords' : 
       #     {'root': 'vpRcL9QNUFj093yMp',
       #     'user': 'GzUODJoYqwg5jjhXG',
       # },
    }

    def __init__(self, config=None):
        if config:
            self.config = config
        self.protocol_name = "ftp"
        self.user = ""
        self.password = ""
        self.authenticated = False

    def connection_made(self, transport) -> None:
        self.transport = transport

        self.logger.log(self.protocol_name + "." + self.logger.CONNECTION, self.transport)
       
        self.transport.write(b'220 ' + bytes(self.config['banner'], "utf-8") + b'\r\n')

    def data_received(self, data):
        # log all data received : print(binascii.hexlify(data))
      
        self.logger.log(self.protocol_name + "." + self.logger.DATA, self.transport, data=data)
        try:
            cmd = data[:4].decode()
            client_input = data[5:].decode().strip('\r\n')
        except UnicodeDecodeError:
            return
    
        if cmd == 'USER':
            self.user = client_input
            self.transport.write("331 User {} OK. Password required\r\n".format(client_input).encode())
        elif cmd == 'PASS':
            self.handle_password(client_input)
        elif cmd == 'SYST':
            if "microsoft" in self.config['banner'].lower():
                self.transport.write(b"215 Windows_NT\r\n")
            else:
                self.transport.write(b"215 UNIX Type: L8\r\n")
        elif cmd == 'QUIT':
            self.transport.write(b"221 Logout\r\n")
            self.connection_lost(None)
        else:
            self.transport.write(b"500 Unknown Command\r\n")

    def handle_password(self, data):
        self.password = data
        
        if not self.user:
            self.transport.write(b"503 Login with USER first.\r\n")
        elif self.user:
            self.logger.log(self.protocol_name + "." + self.logger.LOGIN, self.transport, extra={"username": str(self.user), "password": str(self.password)})

            self.transport.write(b"530 Authentication Failed\r\n")

            #try:
            #    true_password = self.config.get('passwords').get(self.user)
            #    if true_password == self.password:
            #        self.transport.write(b"230 OK. Current restricted directory is /\r\n")
            #        self.authenticated = True    
            #except KeyError:
            #    self.transport.write(b"530 Authentication Failed\r\n")

class FtpHoneypot(BaseHoneypot):
    """common class to all trapster instance"""

    def __init__(self, config, logger, bindaddr="0.0.0.0"):
        super().__init__(config, logger, bindaddr)
        self.handler = lambda: FtpProtocol(config=config)
        self.handler.logger = logger
        self.handler.config = config



from datetime import datetime, timezone
import asyncio, httpx, redis, json, binascii

class BaseLogger(object):
    CONNECTION  = "connection"
    DATA        = "data"
    LOGIN       = "login"
    QUERY       = "query"
    
    def __init__(self, node_id):
        self.node_id = node_id
        self.whitelist_ips = []

    def parse_log(self, logtype, transport, data='', extra={}):
        try:
            dst_ip, dst_port = transport.get_extra_info('sockname')
        except:
            dst_ip = extra.pop("dst_ip", "")
            dst_port = extra.pop("dst_port", 1)
            
        try:
            src_ip, src_port = transport.get_extra_info('peername')
        except:
            # except for the modules : samba, dns, kerberos
            # src_ip is in extra because there is no transport
            src_ip = extra.pop("src_ip", "0.0.0.0")
            src_port = extra.pop("src_port", 1)

        if src_ip in self.whitelist_ips or dst_ip == "255.255.255.255":
            # ignore because IP in whitelist
            return
        else:
            if data:
                # Convert data in Hex format
                data = binascii.hexlify(data).decode()

            event = {
                "device": self.node_id,
                "logtype": logtype,
                "dst_ip": dst_ip,
                "dst_port": dst_port,
                "src_ip": src_ip,
                "src_port": src_port,
                "timestamp": datetime.now(timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f'),
                "data": data,
                "extra": extra
            }

            return event
        
    def log(self, logtype, transport, data='', extra={}):
        event = self.parse_log(logtype, transport, data, extra)
        return event

class JsonLogger(BaseLogger):
    def log(self, logtype, transport, data='', extra={}):
        event = self.parse_log(logtype, transport, data, extra)
            
        if event:
            print(event)

class FileLogger(BaseLogger):
    def __init__(self, node_id, logfile, mode = "w+"):
        self.node_id = node_id
        self.logfile = logfile
        self.mode = mode
        self.file = open(self.logfile, self.mode)

    def log(self, logtype, transport, data='', extra={}):
        event = self.parse_log(logtype, transport, data, extra)
            
        try:
            json.dump(event, self.file)
            self.file.write("\n")
            self.file.flush()
        except IOError as e:
            print(f"An error occurred while writing to the log file: {e}")
        
    def __del__(self):
        if self.file:
            self.file.close()

class RedisLogger(BaseLogger):
    def __init__(self, node_id, host="localhost", port=6379):
        self.node_id = node_id
        self.whitelist_ips = []
        self.r = redis.Redis(host=host, port=port)

    def log(self, logtype, transport, data='', extra={}):
        event = self.parse_log(logtype, transport, data, extra)
        
        if event:
            # Add the event to the sorted set using its timestamp as the primary score
            score = datetime.fromisoformat(event['timestamp']).timestamp()
            self.r.zadd('events', {json.dumps(event): score })

class ApiLogger(BaseLogger):

    def __init__(self, node_id, api_key):
        self.node_id = node_id
        self.headers = {'Authorization': f'token {api_key}'}
        self.whitelist_ips = []

    def log(self, logtype, transport, data='', extra={}):
        event = self.parse_log(logtype, transport, data, extra)
            
        if event:
            loop = asyncio.get_event_loop()            
            loop.create_task(self.post_request(event))

    async def post_request(self, event):
        print(event)
        print('')
        response = await asyncio.to_thread(self._post_request_threaded, event)
        return response

    def _post_request_threaded(self, event):
        with httpx.Client(headers=self.headers) as client:
            response = client.post('http://127.0.0.1:8000/api/v1/event/', json=event, timeout=10)
        return response



from .base import BaseProtocol, BaseHoneypot

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend

import asyncio, asyncssh, os, datetime

async def handle_client(process: asyncssh.SSHServerProcess) -> None:
    """
    Not used, but can be used if we want to allow SSH connection with a very simple simulated prompt
    """
    
    welcome_message = '''Welcome to Ubuntu 20.10 (GNU/Linux 5.8.0-63-generic x86_64)

 * Documentation:  https://help.ubuntu.com
 * Management:     https://landscape.canonical.com
 * Support:        https://ubuntu.com/advantage

  System information as of {time}

  System load:  1.11              Processes:             132
  Usage of /:   50.7% of 1.79TB   Users logged in:       1
  Memory usage: 10%               IPv4 address for eno0: 151.80.38.22
  Swap usage:   6%                IPv6 address for eno0: 2001:41d0:e:b16::1

 * Super-optimized for small spaces - read how we shrank the memory
   footprint of MicroK8s to make it the smallest full K8s around.

   https://ubuntu.com/blog/microk8s-memory-optimisation

0 updates can be installed immediately.
0 of these updates are security updates.

Failed to connect to https://changelogs.ubuntu.com/meta-release. Check your Internet connection or proxy settings


Last login: Wed Jun  8 22:06:15 2022 from 188.64.246.56
'''
    now = datetime.datetime.now(datetime.UTC)
    process.stdout.write(welcome_message.format(time=now.strftime('%a %b  %d %H:%M:%S UTC %Y')))

    while not process.stdin.at_eof():
        try:
            process.stdout.write('root@computer:~# ')
            await process.stdin.readline()
        except asyncssh.misc.BreakReceived:
            process.stdout.write('\n')
            process.stdin.feed_eof()
            process.close()
        except asyncssh.misc.TerminalSizeChanged:
            pass

class SshProtocol(asyncssh.SSHServer, BaseProtocol):

    def __init__(self):
        self.protocol_name = "ssh"
        self.passwords = {'root': '123456',
             'guest': 'password'
            }

    def connection_made(self, transport: asyncssh.SSHServerConnection) -> None:
        self.transport = transport
        self.logger.log(self.protocol_name + "." + self.logger.CONNECTION, self.transport)

    def begin_auth(self, username: str) -> bool:
        # If the user's password is the empty string, no auth is required
        # return passwords.get(username) != ''
        return True

    def password_auth_supported(self) -> bool:
        return True

    async def validate_password(self, username: str, password: str) -> bool:
        self.logger.log(self.protocol_name + "." + self.logger.LOGIN, self.transport, extra={"username":username, "password":password})
        return False
        #return self.passwords.get(username) == password

    def public_key_auth_supported(self) -> bool:
        return True
    
    async def validate_public_key(self, username, key):
        key_type = key.get_algorithm()
        key_data = key.export_public_key().decode()
        fingerprint = key.get_fingerprint()
        
        self.logger.log(self.protocol_name + "." + self.logger.LOGIN, self.transport, extra={
            "username":username, 
            "key_type":key_type, 
            "key_data":key_data,
            "fingerprint": fingerprint
        })
        return False

    def kbdint_auth_supported(self) -> bool:
        return False

    # from https://asyncssh.readthedocs.io/en/latest/_modules/asyncssh/connection.html
    def _send_version(self) -> None:
        """Start the SSH handshake"""
        version = b'SSH-2.0-' + b"OpenSSH_8.3p1 Ubuntu 1ubuntu0.1"

        if self.is_client():
            self._client_version = version
            self.set_extra_info(client_version=version.decode('ascii'))
        else:
            self._server_version = version
            self.set_extra_info(server_version=version.decode('ascii'))

        self._send(version + b'\r\n')


    asyncssh.connection.SSHConnection._send_version = _send_version

class SshHoneypot(BaseHoneypot):
    """common class to all trapster instance"""

    def __init__(self, config, logger, bindaddr="0.0.0.0"):
        super().__init__(config, logger, bindaddr)
        self.handler = SshProtocol
        self.handler.logger = logger
        self.handler.config = config

        self.generate_keys()

    async def _start_server(self):
        try:
            self.server = await asyncssh.create_server(SshProtocol, self.bindaddr, self.port,
                                 server_host_keys=[os.path.dirname(__file__)+"/../data/ssh/ssh_host_key"],
                                 process_factory=handle_client
                                 )
            await self.server.serve_forever()
        except asyncio.CancelledError:
            raise
        except Exception as e:
            print(e)
            return False

    def generate_keys(self):
        # Generate an RSA private key
        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend()
        )

        # Serialize the private key in OpenSSH format
        private_key_openssh = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.OpenSSH,
            encryption_algorithm=serialization.NoEncryption()
        )

        # Serialize the public key in OpenSSH format
        public_key_openssh = private_key.public_key().public_bytes(
            encoding=serialization.Encoding.OpenSSH,
            format=serialization.PublicFormat.OpenSSH
        )

        if not os.path.exists(os.path.dirname(__file__)+"/../data/ssh"):
            os.makedirs(os.path.dirname(__file__)+"/../data/ssh")

        # Save the private and public keys in OpenSSH format
        with open(os.path.dirname(__file__)+"/../data/ssh/ssh_host_key", 'w+') as private:
            private.write(private_key_openssh.decode('utf-8'))

        with open(os.path.dirname(__file__)+"/../data/ssh/ssh_host_key.pub", 'w+') as public:
            public.write(public_key_openssh.decode('utf-8'))
    



# -*- coding: utf-8 -*-

from .base import BaseProtocol, BaseHoneypot

# Based on https://github.com/thomwiggers/httpserver
# __author__ = 'Thom Wiggers, Luuk Scholten'
# __email__ = 'thom@thomwiggers.nl, info@luukscholten.com'
# __version__ = '1.1.0'

import asyncio
import hashlib
import mimetypes
import os
import re
from datetime import datetime, timezone
from http.client import responses
from pathlib import Path


class InvalidRequestError(Exception):
    """Raised for invalid requests. Contains the error code.

    This exception can be transformed to a http response.
    """

    def __init__(self, code, version="HTTP/1.1", body=None, headers=None, *args, **kwargs):
        """Configures a new InvalidRequestError

        Arguments:
            code -- the HTTP error code
        """
        super(InvalidRequestError, self).__init__(*args, **kwargs)
        self.code = code
        self.version = version
        self.body = body
        self.headers = headers if headers is not None else {
            "Content-Type": "text/html; charset=utf-8"
        }

    def get_http_response(self):
        """Get this exception as an HTTP response suitable for output"""
        return HttpProtocol()._get_response(
            version=self.version,
            code=self.code,
            body = self.body,
            headers = self.headers
        )

class HttpProtocol(BaseProtocol):
    """HTTP/1.1 Protocol implementation

    Per connection made, one of these gets instantiated
    """
    config = {
        "basic_auth": False,
        "skin": "defaultApache"
    }

    def __init__(self, config=None, event_loop=None, timeout=10):
        if config:
            self.config = config

        self.protocol_name = "http"
        self.headers = {}

        self._loop = event_loop or asyncio.get_event_loop()
        self._timeout = timeout
        self._timeout_handle = None

    def connection_made(self, transport):
        """Called when the connection is made"""
        self.transport = transport
        # too verbose self.logger.log(self.logger.LOG_HTTP_CONNECTION_MADE, self.transport)
        self.transport = transport
        self.keepalive = True

        if self._timeout:
            self._timout_handle = self._loop.call_later(
                self._timeout, self._handle_timeout)

    def data_received(self, data):
        """Process received data from the socket

        Called when we receive data
        """
        try:
            request = self._parse_headers(data)
            self._handle_request(request)
        except InvalidRequestError as e:
            self._write_response(e.get_http_response())

        if not self.keepalive:
            if self._timeout_handle:
                self._timeout_handle.cancel()
            self.transport.close()

        if self._timeout and self._timeout_handle:
            self._timeout_handle.cancel()
            self._timout_handle = self._loop.call_later(
                self._timeout, self._handle_timeout)

    def _get_response(self, **kwargs):
        """Get a template response

        Use kwargs to add things to the dictionary
        """
        if 'code' not in kwargs:
            kwargs['code'] = 200
        if 'headers' not in kwargs:
            kwargs['headers'] = dict()
        if 'version' not in kwargs:
            kwargs['version'] = 'HTTP/1.1'

        return dict(**kwargs)

    def _write_transport(self, string):
        """Convenience function to write to the transport"""
        if isinstance(string, str):  # we need to convert to bytes
            self.transport.write(string.encode('utf-8'))
        else:
            self.transport.write(string)

    def _write_response(self, response):
        """Write the response back to the client

        Arguments:
        response -- the dictionary containing the response.
        """
        # add custom headers
        if self.config['skin'] == "nasLogin":
            response['headers'].update({
                'Vary': 'Accept-Encoding',
                'Cache-control': 'no-store',
                'X-Content-Type-Options': 'nosniff',
                'X-XSS-Protection': '1; mode=block',
                'X-Frame-Options': 'SAMEORIGIN',
                'P3P': 'CP="IDC DSP COR ADM DEVi TAIi PSA PSD IVAi IVDi CONi HIS OUR IND CNT"',
                'Content-Security-Policy': "base-uri 'self';  connect-src ws: wss: *; default-src 'self' 'unsafe-eval' data: blob: https://*.synology.com https://www.synology.cn/; font-src 'self' data: https://*.googleapis.com https://*.gstatic.com; form-action 'self'; frame-ancestors 'self' https://gofile.me http://gofile.me; frame-src 'self' data: blob: https://*.synology.com https://www.synology.cn/; img-src 'self' data: blob: https://*.google.com https://*.googleapis.com http://*.googlecode.com https://*.gstatic.com; media-src 'self' data: about:;  script-src 'self' 'unsafe-eval' data: blob: https://*.synology.com https://www.synology.cn/ https://*.google.com https://*.googleapis.com; style-src 'self' 'unsafe-inline' https://*.googleapis.com;",
                'Set-Cookie': 'id=;expires=Thu, 01-Jan-1970 00:00:01 GMT;path=/'
            })
        elif self.config['skin'] == "defaultIIS":
            response['headers'].update({
                'Server' : 'Microsoft-IIS/8.5',
                'X-Powered-By': 'ASP.NET',
                'X-AspNet-Version': '4.0.30319',
                'Strict-Transport-Security': 'max-age=31536000; includeSubDomains;',
                'X-XSS-Protection': '1; mode=block',
            })
                

        # write response
        status = '{} {} {}\r\n'.format(response['version'],
                                       response['code'],
                                       responses[response['code']])
        self._write_transport(status)

        if 'body' in response and 'Content-Length' not in response['headers']:
            response['headers']['Content-Length'] = len(response['body'])

        response['headers']['Date'] = datetime.now(timezone.utc).strftime(
            "%a, %d %b %Y %H:%M:%S +0000")

        for (header, content) in response['headers'].items():
            self._write_transport('{}: {}\r\n'.format(header, content))

        self._write_transport('\r\n')
        if 'body' in response:
            self._write_transport(response['body'])


    def _parse_headers(self, data):
        request = dict()

        try:
            request_strings = list(map(lambda x: x.decode(),
                                   data.split(b'\r\n')))
        except UnicodeDecodeError:
            self.keepalive = False  # We don't trust you
            raise InvalidRequestError(400, body='Bad request')

        # Parse request method and HTTP version
        method_line = request_strings[0].split()

        # The first line has either 3 or 2 arguments
        if not (2 <= len(method_line) <= 3):
            # Got an invalid http header
            self.keepalive = False  # We don't trust you
            raise InvalidRequestError(400, body='Bad request')
        # HTTP 0.9 isn't supported.
        if len(method_line) == 2:
            # Got a HTTP/0.9 request
            self.keepalive = False  # HTTP/0.9 won't support persistence
            raise InvalidRequestError(505, body="This server only supports HTTP/1.0"
                                           "and HTTP/1.1")
        else:
            request['version'] = method_line[2]

        # method
        request['method'] = method_line[0]
        # request URI
        request['target'] = method_line[1]

        # Parse the headers
        request["headers"] = {}
        for line in request_strings[1:]:
            if line == '':  # an empty line signals the end of the headers
                break
            header, value = line.split(': ', 1)
            request["headers"][header] = value
        
        # Retrieve POST data
        try:
            post_data = data.split(b'\r\n\r\n')[1]
        except:
            post_data = b'error'

        extra = {"method":request.get('method'), "basic": False, "version" : request.get('version'), "target":request.get('target'), "headers": request.get('headers')}
        if post_data != b'':
            extra["data"] = post_data.decode(errors='backslashreplace')

        # If Authorization Header : basic auth is True
        if request["headers"].get("Authorization"):
            extra['basic'] = True

        # Log request
        if extra['basic'] == True or extra['method'] == 'POST':
            self.logger.log(self.protocol_name + "." + self.logger.LOGIN, self.transport, extra=extra)
        else:
            self.logger.log(self.protocol_name + "." + self.logger.QUERY, self.transport, extra=extra)
 
        return request

    def _get_request_uri(self, request):
        """Parse the request URI into something useful

        Server MUST accept full URIs (5.1.2)"""
        request_uri = request.get('target')
        if request_uri.startswith('/'):  # eg. GET /index.html
            return (request.get('Host', 'localhost').split(':')[0],
                    request_uri[1:])
        elif '://' in request_uri:  # eg. GET http://rded.nl
            locator = request_uri.split('://', 1)[1]
            try:
                host, path = locator.split('/', 1)
            except ValueError:
                host = locator
                path = '/'
            return (host.split(':')[0], path)

    def _handle_request(self, request):
        """Process the headers and get the file"""

        # Check if this is a persistent connection.
        if request.get('version') == 'HTTP/1.1':
            self.keepalive = not request.get('Connection') == 'close'
        elif request.get('version') == 'HTTP/1.0':
            self.keepalive = request.get('Connection') == 'Keep-Alive'

        # Check if we're getting a sane request
        if request.get('method') not in ('GET', 'POST', 'HEAD'):
            raise InvalidRequestError(501, version=request.get('version'), body='Method not implemented')
        if request.get('version') not in ('HTTP/1.0', 'HTTP/1.1'):
            raise InvalidRequestError(505, version=request.get('version'), body='Version not supported. Supported versions are: {}, {}'
                .format('HTTP/1.0', 'HTTP/1.1'))

        host, location = self._get_request_uri(request)

        # We must ignore the Host header if a host is specified in GET
        if host is None:
            host = request.get('Host')


        if self.config.get('directory'):
            script_dir = Path(self.config.get('directory'))
        else:
            script_dir = Path(__file__).resolve().parent / "resources/httpskins"
        
        folder = script_dir / self.config['skin']

        # Ensure the resolved path is within the base directory
        if not script_dir in folder.resolve().parents:
            raise ValueError("Invalid file path")
        
        # Send 401 Unauthorized if basic_auth configuration
        if self.config['basic_auth'] == True:
            with open(folder / "401.html", 'rb') as fp:
                raise InvalidRequestError(401, version=request.get('version'),
                    headers={
                        'WWW-Authenticate': 'Basic realm="Unauthorized"',
                        "Content-Type": "text/html; charset=utf-8"
                    },
                    body=fp.read())

        # Check requested filename
        filename = folder / location

        if str(folder) not in str(filename):
            with open(folder / "404.html", 'rb') as fp:
                raise InvalidRequestError(404, version=request['version'], body=fp.read())

        if os.path.isdir(filename):
            filename = filename / 'index.html'

        if not os.path.isfile(filename):
            with open(folder / "404.html", 'rb') as fp:
                raise InvalidRequestError(404, version=request.get('version'), body=fp.read())

        # Start response with version
        response = self._get_response(version=request.get('version'))

        # timeout negotiation
        match = re.match(r'timeout=(\d+)', request.get('Keep-Alive', ''))
        if match is not None:
            requested_timeout = int(match.group(1))
            if requested_timeout < self._timeout:
                self._timeout = requested_timeout

        # tell the client our timeout
        if self.keepalive:
            response['headers'][
                'Keep-Alive'] = 'timeout={}'.format(self._timeout)

        # Set Content-Type
        response['headers']['Content-Type'] = mimetypes.guess_type(
            filename)[0] or 'text/html'

        # Generate E-tag
        sha1 = hashlib.sha1()
        with open(filename, 'rb') as fp:
            response['body'] = fp.read()
            sha1.update(response['body'])
        etag = sha1.hexdigest()

        # Create 304 response if if-none-match matches etag
        if request.get('If-None-Match') == '"{}"'.format(etag):
            # 304 responses shouldn't contain many headers we might already
            # have added.
            response = self._get_response(code=304)

        response['headers']['Etag'] = '"{}"'.format(etag)

        self._write_response(response)

    def _handle_timeout(self):
        """Handle a timeout"""
        self.transport.close()



class HttpHoneypot(BaseHoneypot):
    """common class to all trapster instance"""

    def __init__(self, config, logger, bindaddr="0.0.0.0"):
        super().__init__(config, logger, bindaddr)
        self.handler = lambda: HttpProtocol(config=config)
        self.handler.logger = logger
        self.handler.config = config
